/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "kernel.h"
#include "frmmain.h"
#include "frmview.h"
#include "frmgroup.h"
#include "frmtexpage.h"
#include "frmframe.h"
#include "frmsequence.h"
#include "glView.h"
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qbuttongroup.h>
#include <qtooltip.h>
#include <qworkspace.h>
#include <qpixmap.h>
#include <qstatusbar.h>

extern frmMain *gbl_frmMain;
std::vector <QPixmap> gbl_pixmaps;

ViewData::ViewData()
{
	axis_aligned = true;
	edit_window = true;
	show_grid = true;
	show_rulers = true;
	use_wireframe = true;
	use_lighting = false;
	use_texture = false;
	
	ortho_width = 100.0f;
	ortho_height = 100.0f;
	
	do_cam_scale = false;
	
	// Metric would be better (scientifically) but
	// imperial is more convinient and easier for the
	// computer due to power of 2 representation.
	small_grid_res = 1.0f;	// 1/8 inch
	medium_grid_res = 8.0f;	// 1 inch
	large_grid_res = 96.0f;	// 1 foot (12 inches)
	
	frm = 0;
}


Model::Model()
{
	snap_to_grid = true;
	modified = false;
	tri_build_count = 0;
	
	sel_vertex = -1;
	sel_triangle = -1;
	
	sel_verticies_count = 0;
	sel_verticies = 0;

	sel_bone = -1;
	sel_joint = -1;

	show_bone_cursor = false;
	do_bone_bind = false;

	frmGrp = new frmGroup( gbl_frmMain->wrkspc, 0, 0);
	frmGrp->mdl = this;

	frmTex = new frmTexPage( gbl_frmMain->wrkspc, 0, 0);
	frmTex->create_views( 0);
	frmTex->mdl = this;

	frmSkel = new frmFrame( gbl_frmMain->wrkspc, 0, 0);
	frmSkel->mdl = this;

	frmSeq = new frmSequence( gbl_frmMain->wrkspc, 0, 0);
	frmSeq->mdl = this;

	krn_add_model_to_editor( this);
}

Model::~Model()
{
	clear();
	delete_all_view_windows();
	delete frmGrp;
	delete frmTex;
	delete frmSkel;
	delete frmSeq;
	krn_remove_model_from_editor( this);
}

void Model::clear()
{
	clear_verticies_selection();
}

void Model::clear_verticies_selection()
{
	// Clear any current selection.
	sel_verticies_count = 0;
	if( sel_verticies) {
		delete [] sel_verticies;
		sel_verticies = 0;
	}
}

void Model::set_caption_for_all_views( const char *s)
{
	for( int i = 0; i < int( view.size()); i++)
		(view[i])->setCaption( s);
}

void Model::delete_all_view_windows()
{
	m_selecting_view = 0;

	// Check if the model should be saved first before deleting
	// all view windows. This is because the model will be destroyed
	// when the last view window is destroyed.
	if( !(view.empty())) {
		if( !((view[0])->save_model_before_destroy())) return;
	}

	// NOTE: Send a close request to each owned view, but don't
	//	really delete it. The view windows will delete their
	//	own memory and the last view window will delete this
	//	model.
	for( int i = 0; i < int( view.size()); i++)
		if( view[i]) emit (view[i])->close();
}

void Model::create_four_view()
{
	// FIXME: Should do error checking, otherwise an infinite loop
	//	may result.
	while( view.size() > 4) destroy_view( view[ view.size() - 1]);
	// XXX: There will ALWAYS be at least one view.
	while( view.size() < 4) create_view_window(
		(QWidget *)(gbl_frmMain->wrkspc));

	show_all_view_windows();
	(view[0])->on_camera_set_front_view();
	(view[1])->on_camera_set_side_view();
	(view[2])->on_camera_set_top_view();

	(view[3])->on_camera_set_front_view();
	ViewData &vd = *((ViewData *)(view[3])->glv->edata);
	vd.edit_window = false;
	vd.use_wireframe = false;
	vd.use_lighting = true;
	vd.use_texture = false;
	krn_tile_windows();
}

frmView *Model::create_view_window( QWidget *p_parent)
{
	frmView *v = new frmView( p_parent, 0, Qt::WDestructiveClose);
	v->create_glview( frmTex->glvTex);
	view.push_back( v);
	v->glv->setFocusPolicy( QWidget::StrongFocus);
	
	if( fname.isEmpty()) v->setCaption( "Untitled");
	else v->setCaption( fname);
	
	v->mdl = this;
	ViewData *vd = new ViewData();
	vd->frm = v;
	v->glv->edata = vd;
	v->glv->on_initialize = vfn_initializeGL;
	v->glv->on_destructor = vfn_destructorGL;
	v->glv->on_resize = vfn_resizeGL;
	v->glv->on_paint = vfn_paintGL;
	v->glv->on_mouse_down = vfn_mouse_down;
	v->glv->on_mouse_move = vfn_mouse_move;
	v->glv->on_mouse_up = vfn_mouse_up;
	v->glv->on_key_down = vfn_key_down;
	
	v->init_popup_menus();
	return v;
}

void Model::destroy_view( frmView *v)
{
	// NOTE: "v" will release its own memory, so "destroy_view" does
	//	not itself release the memory.
	if( v == m_selecting_view) m_selecting_view = 0;
	for( int i = 0; i < int( view.size()); i++) {
		if( v == (view[i])) {
			view.erase( view.begin() + i);
			return;
		}
	}
}

void Model::set_selecting_view( frmView *v)
{
	if( !v) { m_selecting_view = 0; return;}
	for( int i = 0; i < int( view.size()); i++) {
		if( v == (view[i])) { m_selecting_view = v; return;}
	}
	return;
}

void Model::add_vertex( frmView *p_view, const rgVector v)
{
	mdl.add_vertex( krn_snap_to_grid(
		*((ViewData *)p_view->glv->edata), v));
	mdl.transform_skel_frame();
}

void Model::show_all_view_windows() const
{
	for( int i = 0; i < int( view.size()); i++)
		if( !((view[i])->isVisible())) (view[i])->show();
}

void Model::update_all_views() const
{
	for( int i = 0; i < int( view.size()); i++)
		(view[i])->glv->updateGL();
}

int krn_get_prev_edit_mode()
{
	return gbl_frmMain->prev_edMode;
}

int krn_get_edit_mode()
{
	return gbl_frmMain->edMode;
}

Model *krn_create_new_model( std::vector <Model *> &mdl)
{
	Model *m = new Model;
	mdl.push_back( m);
	return m;
}

void krn_delete_model( Model *m)
{
	std::vector <Model *> &mdl = gbl_frmMain->mdl;
	
	if( !m) return;
	for( int i = 0; i < int( mdl.size()); i++) {
		if( (mdl[i]) == m) {
			delete (mdl[i]);
			mdl.erase( mdl.begin() + i);
			return;
		}
	}
}

void krn_create_main_tool_buttons( frmMain *frm)
{
	// Create the mode buttons
	QButtonGroup *g = new QButtonGroup( frm->tbarMain, 0);
	QHBoxLayout *l = new QHBoxLayout( g);
	g->setExclusive( true);
	
	QToolButton *t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_camera.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		      SLOT( on_camera_translate_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Translate Camera");
	
	/*
	t = new QToolButton( g, 0);
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_camera_rotate_Y_mode()));
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Rotate Camera (Y Axis)");
	
	t = new QToolButton( g, 0);
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_camera_rotate_X_mode()));
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Rotate Camera (X Axis)");
	
	t = new QToolButton( g, 0);
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_camera_rotate_Z_mode()));
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Rotate Camera (Z Axis)");
	*/
	//t = new QToolButton( g, 0);
	//frm->connect( t, SIGNAL( toggled( bool)), frm,
	//	SLOT( on_camera_scale_mode()));
	//t->setToggleButton( true);
	//t->setMinimumSize( QSize( 32, 32));
	//l->add( t);
	//QToolTip::add( t, "Scale Camera View");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_select_point.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		      SLOT( on_select_vertex_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Select Vertices");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_add_point.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		      SLOT( on_add_vertex_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Add Vertices");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_select_triangle.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_select_triangle_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Select Triangle");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_add_triangle.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_add_triangle_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Add Triangle");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_select_bone.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_select_bone_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Select Bone");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_add_bone.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_add_bone_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Add Bone");

	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_select_brush.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_select_brush_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Select Brush");
	
	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_add_brush.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_add_brush_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Add Brush");

	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_scale_brush.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_scale_brush_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Scale Brush");

	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_rotate_brush.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		SLOT( on_rotate_brush_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Rotate Brush");

	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_rotate.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		      SLOT( on_rotate_selection_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Rotate Selection");

	t = new QToolButton( g, 0);
	t->setIconSet( QIconSet( QPixmap("./system/icon_scale.bmp") ));
	frm->connect( t, SIGNAL( toggled( bool)), frm,
		      SLOT( on_scale_selection_mode()));
	t->setAutoRaise( true);
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Scale Selection");
	
/*
	t = new QToolButton( g, 0);
	//frm->connect( t, SIGNAL( toggled( bool)), frm,
	//SLOT( on_camera_mode()));
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Select Brush");
	
	t = new QToolButton( g, 0);
	//frm->connect( t, SIGNAL( toggled( bool)), frm,
	//SLOT( on_camera_mode()));
	t->setToggleButton( true);
	t->setMinimumSize( QSize( 32, 32));
	l->add( t);
	QToolTip::add( t, "Select Sub-Division Surface");
	*/
}

// NOTE: "p" MUST be in world space.
int krn_select_vertex( Model &mdl, rgVector p, const rgVector n,
	const float max_dis)
{
	float near_dis = fabs( max_dis);
	int vid = -1;
	const int istop = int( mdl.mdl.v.size());
	p = rgMakePlanar( p, n);
	for( int i = 0; i < istop; i++) {
		
		if( !(mdl.mdl.v[i].hidden())) {
			// Get the current vertex and project it onto the
			// plane of selection.
			const float d = (rgMakePlanar(
				mdl.mdl.get_tv_vertex( i), n) - p).length();
			if( d < near_dis) {
				vid = i;
				near_dis = d;
			}
		}
	}

	return vid;
}

void krn_select_vertices( Model &mdl, const rgVector n)
{
	// The selection box is on the selection plane.
	mdl.clear_verticies_selection();
	std::vector <int> sel;
	const rgVector sb[] = {
		rgMakePlanar( (mdl.sel_box.vertices())[0], n),
		rgMakePlanar( (mdl.sel_box.vertices())[1], n),
		rgMakePlanar( (mdl.sel_box.vertices())[2], n),
		rgMakePlanar( (mdl.sel_box.vertices())[3], n)};
	const int istop = int( mdl.mdl.v.size());
	int i;
	for( i = 0; i < istop; i++) {
		
		if( !(mdl.mdl.v[i].hidden())) {
			// Get the current vertex, project it onto the
			// plane of selection, and check if it is within
			// the selection box.
			if( rgPlanarPointInConvexFace( sb, 4,
				rgMakePlanar( mdl.mdl.get_tv_vertex( i), n)))
				sel.push_back( i);
		}
	}
	
	if( !(sel.size())) return;
	mdl.sel_verticies_count = int( sel.size());
	mdl.sel_verticies = new int[ mdl.sel_verticies_count];
	for( i = 0; i < mdl.sel_verticies_count; i++)
		mdl.sel_verticies[i] = sel[i];
}

void SelectionBox::set_box( const float mdx, const float mdy, const float mcx,
	const float mcy, const GLView *g)
{
	m_v[0] = rgVector( RG_MIN( mdx, mcx), RG_MIN( mdy, mcy), 0);
	m_v[2] = rgVector( RG_MAX( mdx, mcx), RG_MAX( mdy, mcy), 0);
	m_v[1] = rgVector( m_v[2].x, m_v[0].y, 0);
	m_v[3] = rgVector( m_v[0].x, m_v[2].y, 0);
	rggl_ortho_window_to_world( g->width(), g->height(), m_v, 4,
		&(g->vmat.icamM));
}

void SelectionBox::set_box( const rgVector mdp, const rgVector mcp)
{
	m_v[0] = rgVector( RG_MIN( mdp.x, mcp.x), RG_MIN( mdp.y, mcp.y), 0);
	m_v[2] = rgVector( RG_MAX( mdp.x, mcp.x), RG_MAX( mdp.y, mcp.y), 0);
	m_v[1] = rgVector( m_v[2].x, m_v[0].y, 0);
	m_v[3] = rgVector( m_v[0].x, m_v[2].y, 0);
}



void SelectionBox::render( const rgMatrix &m) const
{
	GLint front_and_back[2];
	const GLint &front = front_and_back[0];
	const GLint &back = front_and_back[1];
	glGetIntegerv( GL_POLYGON_MODE, front_and_back);

	const rgMatrix t( m * m_transformation);
	glEnable( GL_BLEND);
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
	glBegin( GL_QUADS);
	glColor4f( 0, 1, 1, 0.125);
	glVertex( t.pointMul( m_v[0]) + m_offset);
	glVertex( t.pointMul( m_v[1]) + m_offset);
	glVertex( t.pointMul( m_v[2]) + m_offset);
	glVertex( t.pointMul( m_v[3]) + m_offset);
	glEnd();
	glDisable( GL_BLEND);

	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
	glBegin( GL_QUADS);
	glColor3f( 0, 0, 1);
	glVertex( t.pointMul( m_v[0]) + m_offset);
	glVertex( t.pointMul( m_v[1]) + m_offset);
	glVertex( t.pointMul( m_v[2]) + m_offset);
	glVertex( t.pointMul( m_v[3]) + m_offset);
	glEnd();

	glPolygonMode( GL_FRONT, front);
	glPolygonMode( GL_BACK, back);
}

frmView *krn_create_model_view_window( Model &mdl)
{
	return mdl.create_view_window( (QWidget *)(gbl_frmMain->wrkspc));
}

void krn_call_open_model()
{
	gbl_frmMain->fileOpen();
}

void krn_call_new_model()
{
	gbl_frmMain->fileNew();
}

void krn_cascade_windows()
{
	gbl_frmMain->wrkspc->cascade();
}

void krn_tile_windows()
{
	gbl_frmMain->wrkspc->tile();
}

inline float hlp_snap_to_grid( const float n, const float sg, const float mg,
			       const float lg)
{
	const float a = ( ( round( n / sg)) * sg);
	const float b = ( ( round( n / mg)) * mg);
	const float c = ( ( round( n / lg)) * lg);
	if( fabs( a - n) < fabs( b - n)) {
		if( fabs( a - n) < fabs( c - n)) return a;
		else return c;
	} else {
		if( fabs( b - n) < fabs( c - n)) return b;
		else return c;
	}
}

rgVector krn_snap_to_grid( ViewData &vd, const rgVector v)
{
	if( !(vd.frm->mdl->snap_to_grid)) return v;
	return rgVector(
		hlp_snap_to_grid( v.x, vd.small_grid_res,
			vd.medium_grid_res, vd.large_grid_res),
		hlp_snap_to_grid( v.y, vd.small_grid_res,
			vd.medium_grid_res, vd.large_grid_res),
		hlp_snap_to_grid( v.z, vd.small_grid_res,
			vd.medium_grid_res, vd.large_grid_res));
}

int krn_load_system_pixmaps()
{
	QPixmap p;
	p.load( "./system/icon_red_h.bmp");
	gbl_pixmaps.push_back( p);
	return 0;
}

void krn_add_model_to_editor( Model *mdl)
{
	if( !mdl) return;
	gbl_frmMain->model.push_back( mdl);
}

void krn_remove_model_from_editor( const Model *mdl)
{
	std::vector <Model *> &m = gbl_frmMain->model;
	const int istop = int( m.size());
	for( int i = 0; i < istop; i++) {
		if( mdl == m[i]) {
			m.erase( m.begin() + i);
			return;
		}
	}
}

void krn_set_selection_center_point( ViewData &vd)
{
	if( vd.frm->mdl->sel_vertices) {
		vd.frm->mdl->selection_center_point =
			vd.frm->mdl->mdl.get_center_of_tv_vertex_cluster(
				vd.frm->mdl->sel_vertices,
				vd.frm->mdl->sel_vertices_count);
	} else {
		vd.frm->mdl->selection_center_point =
			vd.frm->mdl->get_selection_box_center();
	}
}

rgMatrix krn_gen_rotation_matrix( const rgVector p_init, const rgVector p_fini,
	const rgVector center, const rgVector axis, const int ang_lock)
{
	rgMatrix m;
	krn_hlp_gen_rotation_matrix( rgMakePlanar( center, axis), axis,
		p_init, p_fini, ang_lock, m);
	m.translate( -1.0f * center);
	rgMatrix t;
	t.translate( center);
	return (t * m);
}

void krn_set_transformed_selection_box( ViewData &vd)
{
	rgMatrix &m = vd.frm->mdl->tool_matrix;
	vd.frm->mdl->sel_box.transform( m);
	sk3DRomModel &mdl = vd.frm->mdl->mdl;
	int *sb = vd.frm->mdl->sel_vertices;
	const int istop = vd.frm->mdl->sel_vertices_count;
	for( int i = 0; i < istop; i++) {
		mdl.set_tv_vertex( sb[i],
			m.pointMul( mdl.get_tv_vertex( sb[i])));
	}
	mdl.set_vertex_from_tv( sb, istop);
	m.loadIdentity();
}

void krn_hlp_gen_rotation_matrix( const rgVector focus, const rgVector axis,
	const rgVector a, const rgVector b, const int ang_lock, rgMatrix &m)
{
	rgVector u( a - focus);
	rgVector v( b - focus);
	rgFloat ang = rgDotAngle( u, v);
	if( ang_lock > 0) {
		int tmp = abs( int( RG_RADTODEG( ang)));
		tmp = tmp - (tmp % ang_lock);
		if( ang < 0) ang = -RG_DEGTORAD( tmp);
		else ang = RG_DEGTORAD( tmp);
	}
	if( (u.cross( v)).dot( axis) < 0) ang = -ang;
	m.rotate( ang, axis);
}

rgMatrix krn_gen_scale_matrix( const float length, const rgVector win_init,
	const rgVector win_fini, const rgVector center, const float scale)
{
	rgMatrix m;
	rgGLViewMatrix emat;
	emat.do_view_scale( length, win_init, win_fini, scale, &m, 0, true);
	m.translate( -1.0f * center);
	rgMatrix t;
	t.translate( center);
	return (t * m);
}

void krn_print_location( const rgVector p)
{
	QString s;
	s.sprintf( "X: %6.2f, Y: %6.2f, Z: %6.2f", p.x, p.y, p.z);
	(gbl_frmMain->statusBar())->message( s);
}
